"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getRequestedRange = exports.loadPackage = exports.checkInstalled = exports.ensureInstalled = exports.resolveVersion = void 0;
/* Utilities to manage finding, installing and loading legacy versions */
const child_process_1 = require("child_process");
const path = __importStar(require("path"));
const fs_1 = require("fs");
const proper_lockfile_1 = require("proper-lockfile");
const semver = __importStar(require("semver"));
const packageVersion_1 = require("./packageVersion");
// Assuming this file is in dist\test, so go to ..\node_modules\.legacy as the install location
const baseModulePath = path.join(__dirname, "..", "node_modules", ".legacy");
const installedJsonPath = path.join(baseModulePath, "installed.json");
const getModulePath = (version) => path.join(baseModulePath, version);
const resolutionCache = new Map();
// Increment the revision if we want to force installation (e.g. package list changed)
const revision = 1;
async function ensureInstalledJson() {
    if ((0, fs_1.existsSync)(installedJsonPath)) {
        return;
    }
    const release = await (0, proper_lockfile_1.lock)(__dirname, { retries: { forever: true } });
    try {
        // Check it again under the lock
        if ((0, fs_1.existsSync)(installedJsonPath)) {
            return;
        }
        // Create the directory
        (0, fs_1.mkdirSync)(baseModulePath, { recursive: true });
        const data = { revision, installed: [] };
        (0, fs_1.writeFileSync)(installedJsonPath, JSON.stringify(data, undefined, 2), { encoding: "utf8" });
    }
    finally {
        release();
    }
}
function readInstalledJsonNoLock() {
    const data = (0, fs_1.readFileSync)(installedJsonPath, { encoding: "utf8" });
    const installedJson = JSON.parse(data);
    if (installedJson.revision !== revision) {
        // if the revision doesn't match assume that it doesn't match
        return { revision, installed: [] };
    }
    return installedJson;
}
async function readInstalledJson() {
    await ensureInstalledJson();
    const release = await (0, proper_lockfile_1.lock)(installedJsonPath, { retries: { forever: true } });
    try {
        return readInstalledJsonNoLock();
    }
    finally {
        release();
    }
}
const isInstalled = async (version) => (await readInstalledJson()).installed.includes(version);
async function addInstalled(version) {
    await ensureInstalledJson();
    const release = await (0, proper_lockfile_1.lock)(installedJsonPath, { retries: { forever: true } });
    try {
        const installedJson = readInstalledJsonNoLock();
        if (!installedJson.installed.includes(version)) {
            installedJson.installed.push(version);
            (0, fs_1.writeFileSync)(installedJsonPath, JSON.stringify(installedJson, undefined, 2), { encoding: "utf8" });
        }
    }
    finally {
        release();
    }
}
async function removeInstalled(version) {
    await ensureInstalledJson();
    const release = await (0, proper_lockfile_1.lock)(installedJsonPath, { retries: { forever: true } });
    try {
        const installedJson = readInstalledJsonNoLock();
        installedJson.installed = installedJson.installed.filter((value) => value !== version);
        (0, fs_1.writeFileSync)(installedJsonPath, JSON.stringify(installedJson, undefined, 2), { encoding: "utf8" });
    }
    finally {
        release();
    }
}
function resolveVersion(requested, installed) {
    const cachedVersion = resolutionCache.get(requested);
    if (cachedVersion) {
        return cachedVersion;
    }
    if (semver.valid(requested)) {
        // If it is a valid semver already instead of a range, just use it
        resolutionCache.set(requested, requested);
        return requested;
    }
    if (installed) {
        // Check the install directory instad of asking NPM for it.
        const files = (0, fs_1.readdirSync)(baseModulePath, { withFileTypes: true });
        let found;
        files.map((dirent) => {
            if (dirent.isDirectory() && semver.valid(dirent.name) && semver.satisfies(dirent.name, requested)) {
                if (!found || semver.lt(found, dirent.name)) {
                    found = dirent.name;
                }
            }
        });
        if (found) {
            return found;
        }
        throw new Error(`No matching version found in ${baseModulePath}`);
    }
    else {
        let result = (0, child_process_1.execSync)(`npm v @fluidframework/container-loader@"${requested}" version --json`, { encoding: "utf8" });
        // if we are requesting an x.x.0-0 prerelease and failed the first try, try
        // again using the virtualPatch schema
        if (result === "") {
            const requestedVersion = new semver.SemVer(requested.substring(requested.indexOf("^") + 1));
            if (requestedVersion.patch === 0 && requestedVersion.prerelease.length > 0) {
                const retryVersion = `^${requestedVersion.major}.${requestedVersion.minor}.1000-0`;
                result = (0, child_process_1.execSync)(`npm v @fluidframework/container-loader@"${retryVersion}" version --json`, { encoding: "utf8" });
            }
        }
        try {
            const versions = JSON.parse(result);
            const version = Array.isArray(versions) ? versions.sort(semver.rcompare)[0] : versions;
            if (!version) {
                throw new Error(`No version found for ${requested}`);
            }
            resolutionCache.set(requested, version);
            return version;
        }
        catch (e) {
            throw new Error(`Error parsing versions for ${requested}`);
        }
    }
}
exports.resolveVersion = resolveVersion;
async function ensureModulePath(version, modulePath) {
    const release = await (0, proper_lockfile_1.lock)(baseModulePath, { retries: { forever: true } });
    try {
        console.log(`Installing version ${version}`);
        if (!(0, fs_1.existsSync)(modulePath)) {
            // Create the under the baseModulePath lock
            (0, fs_1.mkdirSync)(modulePath, { recursive: true });
        }
    }
    finally {
        release();
    }
}
async function ensureInstalled(requested, packageList, force) {
    if (requested === packageVersion_1.pkgVersion) {
        return;
    }
    const version = resolveVersion(requested, false);
    const modulePath = getModulePath(version);
    if (!force && await isInstalled(version)) {
        return { version, modulePath };
    }
    await ensureModulePath(version, modulePath);
    // Release the __dirname but lock the modulePath so we can do parallel installs
    const release = await (0, proper_lockfile_1.lock)(modulePath, { retries: { forever: true } });
    try {
        if (force) {
            // remove version from install.json under the modulePath lock
            await removeInstalled(version);
        }
        // Check installed status again under lock the modulePath lock
        if (force || !await isInstalled(version)) {
            // Install the packages
            await new Promise((resolve, reject) => (0, child_process_1.exec)(`npm init --yes`, { cwd: modulePath }, (error, stdout, stderr) => {
                if (error) {
                    reject(new Error(`Failed to initialize install directory ${modulePath}`));
                }
                resolve();
            }));
            await new Promise((resolve, reject) => (0, child_process_1.exec)(`npm i --no-package-lock ${packageList.map((pkg) => `${pkg}@${version}`).join(" ")}`, { cwd: modulePath }, (error, stdout, stderr) => {
                if (error) {
                    reject(new Error(`Failed to install in ${modulePath}\n${stderr}`));
                }
                resolve();
            }));
            // add it to the install.json under the modulePath lock.
            await addInstalled(version);
        }
        return { version, modulePath };
    }
    catch (e) {
        // rmdirSync recursive flags introduced in Node v12.10
        // Remove the `as any` cast once node typing is updated.
        try {
            fs_1.rmdirSync(modulePath, { recursive: true });
        }
        catch (ex) { }
        throw new Error(`Unable to install version ${version}\n${e}`);
    }
    finally {
        release();
    }
}
exports.ensureInstalled = ensureInstalled;
function checkInstalled(requested) {
    const version = resolveVersion(requested, true);
    const modulePath = getModulePath(version);
    if ((0, fs_1.existsSync)(modulePath)) {
        // assume it is valid if it exists
        return { version, modulePath };
    }
    throw new Error(`Requested version ${requested} resolved to ${version} is not installed`);
}
exports.checkInstalled = checkInstalled;
const loadPackage = (modulePath, pkg) => 
// eslint-disable-next-line @typescript-eslint/no-require-imports, @typescript-eslint/no-unsafe-return
require(path.join(modulePath, "node_modules", pkg));
exports.loadPackage = loadPackage;
/**
 * Used to get the major version number above or below the baseVersion.
 * @param baseVersion - The base version to move from (eg. "0.60.0")
 * @param requested - Number representing the major version to move from. These are
 * generally negative to move back versions (eg. -1).
 * Note: If the requested number is a string then that will be the returned value
 */
function getRequestedRange(baseVersion, requested) {
    if (requested === undefined || requested === 0) {
        return baseVersion;
    }
    if (typeof requested === "string") {
        return requested;
    }
    const version = new semver.SemVer(baseVersion);
    // ask for prerelease in case we just bumped the version and haven't release the previous version yet.
    if (version.major === 1) {
        if (requested === -1) {
            return "^0.59.0-0";
        }
        else if (requested === -2) {
            return "^0.58.0-0";
        }
    }
    return `^${version.major}.${version.minor + requested}.0-0`;
}
exports.getRequestedRange = getRequestedRange;
//# sourceMappingURL=versionUtils.js.map